"""Pywikibot metadata file.

.. versionadded:: 4.0
"""
#
# (C) Pywikibot team, 2020-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from time import strftime


__version__ = '10.7.4'
__url__ = 'https://www.mediawiki.org/wiki/Manual:Pywikibot'
__copyright__ = f'2003-{strftime("%Y")}, Pywikibot team'
